package mvd.jester;

import java.io.IOException;
import java.time.LocalTime;
import java.util.Set;
import java.util.TreeSet;
import java.util.stream.Collectors;
import com.google.common.collect.ImmutableList;
import com.google.common.collect.ObjectArrays;
import com.google.common.collect.Table;
import com.google.common.collect.TreeBasedTable;
import com.google.common.math.DoubleMath;
import org.apache.commons.csv.CSVFormat;
import org.apache.commons.csv.CSVPrinter;
import mvd.jester.model.Task;
import mvd.jester.simulator.AbstractSimulator;
import mvd.jester.tests.AbstractTest;
import mvd.jester.utils.Logger;

/**
 * ResultCollector
 */
public class ResultLogger {

    private final long numberOfProcessors;

    public ResultLogger(long numberOfProcessors) {
        this.numberOfProcessors = numberOfProcessors;
    }

    public <T extends Task> void logAll(Set<ResultCollector<AbstractTest<T>>> testResults,
            Set<ResultCollector<AbstractSimulator>> simResults) {
        logTests(testResults);
        logSimulations(simResults);
    }

    public <T extends Task> void logTests(Set<ResultCollector<AbstractTest<T>>> results) {
        if (!results.isEmpty()) {
            logFeasibility(results, "test");
        }
    }

    public void logSimulations(Set<ResultCollector<AbstractSimulator>> results) {
        if (!results.isEmpty()) {
            logFeasibility(results, "sim");
            logTaskRatio(results, "sim");
        }
    }


    public <T extends TypeInterface> void logFeasibilityLevel(Set<ResultCollector<T>> results,
            String type) {
        // LocalTime date = LocalTime.now();
        // Logger log = new Logger("./results/feasibility_level_" + type + "_" + numberOfProcessors
        // + "_" + date.getHour() + ":" + date.getMinute() + ".txt");

        // Table<Long, ResultCollector<T>, Double> resultTable = TreeBasedTable.create();
        // Set<ResultCollector<T>> resultCollectors = new TreeSet<>();

        // for (long util = 0; util <= numberOfProcessors * 10; util += numberOfProcessors / 4) {
        // for (ResultCollector<T> rc : results) {
        // resultCollectors.add(rc);
        // final long local_util = util;

        // Supplier<Stream<SchedulingInfo>> schedulingResults = () -> rc.getResults().stream()
        // .filter(r -> Math.round(r.getUtilization() * 10 / (numberOfProcessors / 4))
        // * (numberOfProcessors / 4) == local_util);
        // if (schedulingResults.get().filter(r -> !r.checkTasksetFeasible()).count() > 0) {
        // double feasibleTasksets = (double) schedulingResults.get()
        // .filter(r -> r.checkLevelFail(Level.HIGH)).count()
        // / schedulingResults.get().filter(r -> !r.checkTasksetFeasible())
        // .count();
        // resultTable.put(util, rc, feasibleTasksets);
        // } else {
        // resultTable.put(util, rc, 0.);
        // }
        // }
        // }

        // logData(log, resultTable, resultCollectors, "Utilization");
    }



    public <T extends TypeInterface> void logFeasibility(Set<ResultCollector<T>> results,
            String type) {
        LocalTime date = LocalTime.now();
        Logger log = new Logger("./results/feasibility_" + type + "_" + numberOfProcessors + "_"
                + date.getHour() + ":" + date.getMinute() + ".txt");

        Table<Double, ResultCollector<T>, Long> resultTable = TreeBasedTable.create();
        Set<ResultCollector<T>> resultCollectors = new TreeSet<>();

        for (double util = 0.25; util <= numberOfProcessors; util += 0.25) {
            for (ResultCollector<T> rc : results) {
                resultCollectors.add(rc);
                final double local_util = util;
                long feasibleTasksets = rc.getResults().stream()
                        .filter(r -> DoubleMath.fuzzyEquals(r.getUtilization(), local_util, 0.125))
                        .filter(r -> r.checkTasksetFeasible()).count();
                resultTable.put(util, rc, feasibleTasksets);
            }
        }

        logData(log, resultTable, resultCollectors, "Utilization");
    }

    public <T extends TypeInterface> void logFeasibilityRatio(Set<ResultCollector<T>> results,
            String type) {
        // LocalTime date = LocalTime.now();
        // Logger log = new Logger("./results/feasibility_ratio_" + type + "_" + numberOfProcessors
        // + "_" + date.getHour() + ":" + date.getMinute() + ".txt");

        // Table<Long, ResultCollector<T>, Double> resultTable = TreeBasedTable.create();
        // Set<ResultCollector<T>> resultCollectors = new TreeSet<>();

        // for (long util = 0; util <= numberOfProcessors * 10; util += numberOfProcessors / 4) {
        // for (ResultCollector<T> rc : results) {
        // resultCollectors.add(rc);
        // final long local_util = util;
        // Supplier<Stream<SchedulingInfo>> schedulingResults = () -> rc.getResults().stream()
        // .filter(r -> Math.round(r.getUtilization() * 10 / (numberOfProcessors / 4))
        // * (numberOfProcessors / 4) == local_util);
        // if (schedulingResults.get().count() > 0) {
        // double feasibleTasksets =
        // (double) schedulingResults.get().filter(r -> r.checkTasksetFeasible())
        // .count() / schedulingResults.get().count();
        // resultTable.put(util, rc, feasibleTasksets);
        // } else {
        // resultTable.put(util, rc, 1.);
        // }
        // }
        // }

        // logData(log, resultTable, resultCollectors, "Utilization");
    }

    public <T extends TypeInterface> void logTaskRatio(Set<ResultCollector<T>> results,
            String type) {
        // LocalTime date = LocalTime.now();
        // Logger log = new Logger("./results/task_ratio_" + type + "_" + numberOfProcessors + "_"
        // + date.getHour() + ":" + date.getMinute() + ".txt");

        // Table<Long, ResultCollector<T>, Long> resultTable = TreeBasedTable.create();
        // Set<ResultCollector<T>> resultCollectors = new TreeSet<>();

        // for (long ratio = 0; ratio <= 10; ratio += 1) {
        // for (ResultCollector<T> rc : results) {
        // resultCollectors.add(rc);
        // final long local_ratio = ratio;
        // long feasibleTasksets = rc.getResults().stream()
        // .filter(r -> Math.ceil(r.getParallelTaskRatio() * 10) == local_ratio)
        // .filter(r -> r.checkTasksetFeasible()).count();
        // resultTable.put(ratio, rc, feasibleTasksets);
        // }
        // }

        // logData(log, resultTable, resultCollectors, "TaskRatio");
    }

    private <T extends TypeInterface> void logData(Logger log,
            Table<Double, ResultCollector<T>, ? extends Number> resultTable,
            Set<ResultCollector<T>> resultCollectors, String xDataName) {
        final Appendable out = new StringBuilder();
        try {
            String[] resultCollectorNames = resultCollectors.stream()
                    .map(ResultCollector<T>::getName).toArray(String[]::new);
            String[] header = ObjectArrays.concat(xDataName, resultCollectorNames);
            final CSVPrinter printer = CSVFormat.DEFAULT.withHeader(header).print(out);

            printer.printRecords(resultTable.rowMap().entrySet().stream()
                    .map(entry -> ImmutableList.builder().add((double) entry.getKey())
                            .addAll(entry.getValue().values()).build())
                    .collect(Collectors.toList()));
        } catch (final IOException e) {
            e.printStackTrace();
        }

        log.log(out);
        log.finalize();
    }
}
